@@ -21,19 +21,19 @@ use async_recursion::async_recursion;
 use log::{debug, info};
 use std::marker::{Send, Sync};
 use tokio::time::Instant;
-use winter_crypto::{Digest, Hasher};
-
-use serde::{Deserialize, Serialize};
+use winter_crypto::Hasher;
 
 use keyed_priority_queue::{Entry, KeyedPriorityQueue};
+use std::collections::HashSet;
 
 /// The default azks key
 pub const DEFAULT_AZKS_KEY: u8 = 1u8;
 
 /// An append-only zero knowledge set, the data structure used to efficiently implement
 /// a auditable key directory.
-#[derive(Debug, Serialize, Deserialize, Eq, PartialEq)]
-#[serde(bound = "")]
+#[derive(Debug, Eq, PartialEq)]
+#[cfg_attr(feature = "serde", derive(serde::Deserialize, serde::Serialize))]
+#[cfg_attr(feature = "serde", serde(bound = ""))]
 pub struct Azks {
     /// The latest complete epoch
     pub latest_epoch: u64,
@@ -49,15 +49,18 @@ impl Storable for Azks {
     }
 
     fn get_id(&self) -> u8 {
-        1u8
+        DEFAULT_AZKS_KEY
     }
 
     fn get_full_binary_key_id(key: &u8) -> Vec<u8> {
         vec![StorageType::Azks as u8, *key]
     }
 
-    fn key_from_full_binary(_bin: &[u8]) -> Result<u8, String> {
-        Ok(1u8)
+    fn key_from_full_binary(bin: &[u8]) -> Result<u8, String> {
+        if bin.is_empty() || bin[0] != StorageType::Azks as u8 {
+            return Err("Not an AZKS key".to_string());
+        }
+        Ok(DEFAULT_AZKS_KEY)
     }
 }
 
@@ -99,14 +102,18 @@ impl Azks {
         let new_leaf = get_leaf_node::<H, S>(
             storage,
             node.label,
-            node.hash.as_bytes().as_ref(),
+            &node.hash,
             NodeLabel::root(),
             self.latest_epoch,
         )
         .await?;
 
-        let mut root_node =
-            HistoryTreeNode::get_from_storage(storage, NodeKey(NodeLabel::root())).await?;
+        let mut root_node = HistoryTreeNode::get_from_storage(
+            storage,
+            &NodeKey(NodeLabel::root()),
+            self.get_latest_epoch(),
+        )
+        .await?;
         root_node
             .insert_single_leaf::<_, H>(storage, new_leaf, self.latest_epoch, &mut self.num_nodes)
             .await?;
@@ -129,9 +136,6 @@ impl Azks {
         storage: &S,
         insertion_set: &[Node<H>],
     ) -> Result<u64, AkdError> {
-        let mut load_count: u64 = 0;
-        let mut current_nodes = vec![NodeKey(NodeLabel::root())];
-
         let prefixes_set = crate::utils::build_prefixes_set(
             insertion_set
                 .iter()
@@ -140,8 +144,25 @@ impl Azks {
                 .as_ref(),
         );
 
+        self.bfs_preload_nodes::<S, H>(storage, prefixes_set).await
+    }
+
+    /// Preloads given nodes using breadth-first search.
+    pub async fn bfs_preload_nodes<S: Storage + Sync + Send, H: Hasher>(
+        &self,
+        storage: &S,
+        nodes_to_load: HashSet<NodeLabel>,
+    ) -> Result<u64, AkdError> {
+        let mut load_count: u64 = 0;
+        let mut current_nodes = vec![NodeKey(NodeLabel::root())];
+
         while !current_nodes.is_empty() {
-            let nodes = HistoryTreeNode::batch_get_from_storage(storage, current_nodes).await?;
+            let nodes = HistoryTreeNode::batch_get_from_storage(
+                storage,
+                &current_nodes,
+                self.get_latest_epoch(),
+            )
+            .await?;
             load_count += nodes.len() as u64;
 
             current_nodes = Vec::<NodeKey>::new();
@@ -149,16 +170,16 @@ impl Azks {
 
             // This for loop is just getting the keys for states that need to be loaded.
             for node in &nodes {
-                node_states.push(get_state_map_key(node, node.get_latest_epoch()?));
+                node_states.push(get_state_map_key(node, node.get_latest_epoch()));
             }
-            let states = storage.batch_get::<HistoryNodeState>(node_states).await?;
+            let states = storage.batch_get::<HistoryNodeState>(&node_states).await?;
             load_count += states.len() as u64;
 
             // Now that states are loaded in the cache, you can read and access them.
             // Note, the two for loops are needed because otherwise, you'd be accessing remote storage
             // individually for each node's state.
             for node in &nodes {
-                if !prefixes_set.contains(&node.label) {
+                if !nodes_to_load.contains(&node.label) {
                     // Only continue to traverse nodes which are relevant prefixes to insertion_set
                     continue;
                 }
@@ -202,12 +223,15 @@ impl Azks {
         );
 
         self.increment_epoch();
-        self.preload_nodes_for_insertion::<S, H>(storage, &insertion_set)
-            .await?;
+
         let mut hash_q = KeyedPriorityQueue::<NodeLabel, i32>::new();
         let mut priorities: i32 = 0;
-        let mut root_node =
-            HistoryTreeNode::get_from_storage(storage, NodeKey(NodeLabel::root())).await?;
+        let mut root_node = HistoryTreeNode::get_from_storage(
+            storage,
+            &NodeKey(NodeLabel::root()),
+            self.get_latest_epoch(),
+        )
+        .await?;
         for node in insertion_set {
             let new_leaf = if append_only_usage {
                 get_leaf_node_without_hashing::<H, S>(
@@ -221,7 +245,7 @@ impl Azks {
                 get_leaf_node::<H, S>(
                     storage,
                     node.label,
-                    node.hash.as_bytes().as_ref(),
+                    &node.hash,
                     NodeLabel::root(),
                     self.latest_epoch,
                 )
@@ -238,9 +262,14 @@ impl Azks {
             priorities -= 1;
         }
 
+        // Now hash up the tree, the highest priority items will be closer to the leaves.
         while let Some((next_node_label, _)) = hash_q.pop() {
-            let mut next_node: HistoryTreeNode =
-                HistoryTreeNode::get_from_storage(storage, NodeKey(next_node_label)).await?;
+            let mut next_node: HistoryTreeNode = HistoryTreeNode::get_from_storage(
+                storage,
+                &NodeKey(next_node_label),
+                self.get_latest_epoch(),
+            )
+            .await?;
 
             next_node
                 .update_hash::<_, H>(storage, self.latest_epoch)
@@ -288,24 +317,33 @@ impl Azks {
         let (longest_prefix_membership_proof, lcp_node_label) = self
             .get_membership_proof_and_node(storage, label, epoch)
             .await?;
-        let lcp_node: HistoryTreeNode =
-            HistoryTreeNode::get_from_storage(storage, NodeKey(lcp_node_label)).await?;
+        let lcp_node: HistoryTreeNode = HistoryTreeNode::get_from_storage(
+            storage,
+            &NodeKey(lcp_node_label),
+            self.get_latest_epoch(),
+        )
+        .await?;
         let longest_prefix = lcp_node.label;
         // load with placeholder nodes, to be replaced in the loop below
         let mut longest_prefix_children = [Node {
-            label: NodeLabel::root(),
-            hash: crate::utils::empty_node_hash::<H>(),
+            label: EMPTY_LABEL,
+            hash: crate::utils::empty_node_hash_no_label::<H>(),
         }; ARITY];
         let state = lcp_node.get_state_at_epoch(storage, epoch).await?;
-
         for (i, child) in state.child_states.iter().enumerate() {
             match child {
                 None => {
+                    debug!("i = {}, empty", i);
                     continue;
                 }
                 Some(child) => {
-                    let unwrapped_child: HistoryTreeNode =
-                        HistoryTreeNode::get_from_storage(storage, NodeKey(child.label)).await?;
+                    let unwrapped_child: HistoryTreeNode = HistoryTreeNode::get_from_storage(
+                        storage,
+                        &NodeKey(child.label),
+                        self.get_latest_epoch(),
+                    )
+                    .await?;
+                    debug!("Label of child {} is {:?}", i, unwrapped_child.label);
                     longest_prefix_children[i] = Node {
                         label: unwrapped_child.label,
                         hash: unwrapped_child
@@ -315,6 +353,7 @@ impl Azks {
                 }
             }
         }
+        debug!("Lcp label = {:?}", longest_prefix);
         Ok(NonMembershipProof {
             label,
             longest_prefix,
@@ -341,7 +380,12 @@ impl Azks {
         // Suppose the epochs start_epoch and end_epoch exist in the set.
         // This function should return the proof that nothing was removed/changed from the tree
         // between these epochs.
-        let node = HistoryTreeNode::get_from_storage(storage, NodeKey(NodeLabel::root())).await?;
+        let node = HistoryTreeNode::get_from_storage(
+            storage,
+            &NodeKey(NodeLabel::root()),
+            self.get_latest_epoch(),
+        )
+        .await?;
         let (unchanged, leaves) = self
             .get_append_only_proof_helper::<_, H>(storage, node, start_epoch, end_epoch)
             .await?;
@@ -361,7 +405,7 @@ impl Azks {
     ) -> Result<AppendOnlyHelper<H>, AkdError> {
         let mut unchanged = Vec::<Node<H>>::new();
         let mut leaves = Vec::<Node<H>>::new();
-        if node.get_latest_epoch()? <= start_epoch {
+        if node.get_latest_epoch() <= start_epoch {
             if node.is_root() {
                 // this is the case where the root is unchanged since the last epoch
                 return Ok((unchanged, leaves));
@@ -370,7 +414,7 @@ impl Azks {
             unchanged.push(Node::<H> {
                 label: node.label,
                 hash: node
-                    .get_value_without_label_at_epoch::<_, H>(storage, node.get_latest_epoch()?)
+                    .get_value_without_label_at_epoch::<_, H>(storage, node.get_latest_epoch())
                     .await?,
             });
             return Ok((unchanged, leaves));
@@ -383,7 +427,7 @@ impl Azks {
             leaves.push(Node::<H> {
                 label: node.label,
                 hash: node
-                    .get_value_without_label_at_epoch::<_, H>(storage, node.get_latest_epoch()?)
+                    .get_value_without_label_at_epoch::<_, H>(storage, node.get_latest_epoch())
                     .await?,
             });
         } else {
@@ -401,7 +445,8 @@ impl Azks {
                     Some(child_node_state) => {
                         let child_node = HistoryTreeNode::get_from_storage(
                             storage,
-                            NodeKey(child_node_state.label),
+                            &NodeKey(child_node_state.label),
+                            self.get_latest_epoch(),
                         )
                         .await?;
                         let mut rec_output = self
@@ -426,7 +471,7 @@ impl Azks {
     pub async fn get_root_hash<S: Storage + Sync + Send, H: Hasher>(
         &self,
         storage: &S,
-    ) -> Result<H::Digest, HistoryTreeNodeError> {
+    ) -> Result<H::Digest, AkdError> {
         self.get_root_hash_at_epoch::<_, H>(storage, self.get_latest_epoch())
             .await
     }
@@ -438,17 +483,20 @@ impl Azks {
         &self,
         storage: &S,
         epoch: u64,
-    ) -> Result<H::Digest, HistoryTreeNodeError> {
+    ) -> Result<H::Digest, AkdError> {
         if self.latest_epoch < epoch {
             // cannot retrieve information for future epoch
-            return Err(HistoryTreeNodeError::NonexistentAtEpoch(
-                NodeLabel::root(),
-                self.latest_epoch,
+            return Err(AkdError::HistoryTreeNode(
+                HistoryTreeNodeError::NonexistentAtEpoch(NodeLabel::root(), epoch),
             ));
         }
-        let root_node: HistoryTreeNode =
-            HistoryTreeNode::get_from_storage(storage, NodeKey(NodeLabel::root())).await?;
-        root_node.get_value_at_epoch::<_, H>(storage, epoch).await
+        let root_node: HistoryTreeNode = HistoryTreeNode::get_from_storage(
+            storage,
+            &NodeKey(NodeLabel::root()),
+            self.get_latest_epoch(),
+        )
+        .await?;
+        Ok(root_node.get_value_at_epoch::<_, H>(storage, epoch).await?)
     }
 
     /// Gets the latest epoch of this azks. If an update aka epoch transition
@@ -472,8 +520,12 @@ impl Azks {
         epoch: u64,
     ) -> Result<(MembershipProof<H>, NodeLabel), AkdError> {
         let mut layer_proofs = Vec::new();
-        let mut curr_node: HistoryTreeNode =
-            HistoryTreeNode::get_from_storage(storage, NodeKey(NodeLabel::root())).await?;
+        let mut curr_node: HistoryTreeNode = HistoryTreeNode::get_from_storage(
+            storage,
+            &NodeKey(NodeLabel::root()),
+            self.get_latest_epoch(),
+        )
+        .await?;
         let mut dir = curr_node.label.get_dir(label);
         let mut equal = label == curr_node.label;
         let mut prev_node = NodeLabel::root();
@@ -481,15 +533,12 @@ impl Azks {
             prev_node = curr_node.label;
             let curr_state = curr_node.get_state_at_epoch(storage, epoch).await?;
             let mut nodes = [Node::<H> {
-                label: NodeLabel::root(),
-                hash: H::hash(&[0u8]),
+                label: EMPTY_LABEL,
+                hash: H::hash(&EMPTY_VALUE),
             }; ARITY - 1];
             let mut count = 0;
-            let direction = dir.ok_or_else(|| {
-                AkdError::HistoryTreeNode(HistoryTreeNodeError::NoDirection(
-                    curr_node.label.get_val(),
-                    None,
-                ))
+            let direction = dir.ok_or({
+                AkdError::HistoryTreeNode(HistoryTreeNodeError::NoDirection(curr_node.label, None))
             })?;
             let next_state = curr_state.get_child_state_in_dir(direction);
             if next_state == None {
@@ -497,14 +546,14 @@ impl Azks {
             }
             for i in 0..ARITY {
                 let no_direction_error = AkdError::HistoryTreeNode(
-                    HistoryTreeNodeError::NoDirection(curr_node.label.get_val(), None),
+                    HistoryTreeNodeError::NoDirection(curr_node.label, None),
                 );
                 if i != dir.ok_or(no_direction_error)? {
                     nodes[count] = Node::<H> {
                         label: optional_history_child_state_to_label(&curr_state.child_states[i]),
                         hash: to_digest::<H>(&optional_history_child_state_to_hash::<H>(
                             &curr_state.child_states[i],
-                        )?)?,
+                        ))?,
                     };
                     count += 1;
                 }
@@ -516,11 +565,12 @@ impl Azks {
             });
             let new_curr_node: HistoryTreeNode = HistoryTreeNode::get_from_storage(
                 storage,
-                NodeKey(
+                &NodeKey(
                     curr_node
                         .get_child_label_at_epoch::<_, H>(storage, epoch, dir)
                         .await?,
                 ),
+                self.get_latest_epoch(),
             )
             .await?;
             curr_node = new_curr_node;
@@ -528,8 +578,12 @@ impl Azks {
             equal = label == curr_node.label;
         }
         if !equal {
-            let new_curr_node: HistoryTreeNode =
-                HistoryTreeNode::get_from_storage(storage, NodeKey(prev_node)).await?;
+            let new_curr_node: HistoryTreeNode = HistoryTreeNode::get_from_storage(
+                storage,
+                &NodeKey(prev_node),
+                self.get_latest_epoch(),
+            )
+            .await?;
             curr_node = new_curr_node;
 
             layer_proofs.pop();
@@ -697,7 +751,7 @@ mod tests {
         let mut proof = azks
             .get_membership_proof(&db, insertion_set[0].label, 1)
             .await?;
-        let hash_val = Blake3::hash(&[0u8]);
+        let hash_val = Blake3::hash(&EMPTY_VALUE);
         proof = MembershipProof::<Blake3> {
             label: proof.label,
             hash_val,
@@ -706,7 +760,7 @@ mod tests {
         assert!(
             !verify_membership::<Blake3>(azks.get_root_hash::<_, Blake3>(&db).await?, &proof)
                 .is_ok(),
-            "Membership proof does verifies, despite being wrong"
+            "Membership proof does verify, despite being wrong"
         );
 
         Ok(())
@@ -715,31 +769,32 @@ mod tests {
     #[tokio::test]
     async fn test_membership_proof_intermediate() -> Result<(), AkdError> {
         let db = AsyncInMemoryDatabase::new();
+
         let mut insertion_set: Vec<Node<Blake3>> = vec![];
-        insertion_set.push(Node::<Blake3> {
-            label: NodeLabel::new(0b0, 64),
-            hash: Blake3::hash(&[]),
+        insertion_set.push(Node {
+            label: NodeLabel::new(byte_arr_from_u64(0b0), 64),
+            hash: Blake3::hash(&EMPTY_VALUE),
         });
-        insertion_set.push(Node::<Blake3> {
-            label: NodeLabel::new(0b1 << 63, 64),
-            hash: Blake3::hash(&[]),
+        insertion_set.push(Node {
+            label: NodeLabel::new(byte_arr_from_u64(0b1 << 63), 64),
+            hash: Blake3::hash(&EMPTY_VALUE),
         });
-        insertion_set.push(Node::<Blake3> {
-            label: NodeLabel::new(0b11 << 62, 64),
-            hash: Blake3::hash(&[]),
+        insertion_set.push(Node {
+            label: NodeLabel::new(byte_arr_from_u64(0b11 << 62), 64),
+            hash: Blake3::hash(&EMPTY_VALUE),
         });
-        insertion_set.push(Node::<Blake3> {
-            label: NodeLabel::new(0b01 << 62, 64),
-            hash: Blake3::hash(&[]),
+        insertion_set.push(Node {
+            label: NodeLabel::new(byte_arr_from_u64(0b01 << 62), 64),
+            hash: Blake3::hash(&EMPTY_VALUE),
         });
-        insertion_set.push(Node::<Blake3> {
-            label: NodeLabel::new(0b111 << 61, 64),
-            hash: Blake3::hash(&[]),
+        insertion_set.push(Node {
+            label: NodeLabel::new(byte_arr_from_u64(0b111 << 61), 64),
+            hash: Blake3::hash(&EMPTY_VALUE),
         });
         let mut azks = Azks::new::<_, Blake3>(&db).await?;
         azks.batch_insert_leaves::<_, Blake3>(&db, insertion_set)
             .await?;
-        let search_label = NodeLabel::new(0b1111 << 60, 64);
+        let search_label = NodeLabel::new(byte_arr_from_u64(0b1111 << 60), 64);
         let proof = azks.get_non_membership_proof(&db, search_label, 1).await?;
         assert!(
             verify_nonmembership::<Blake3>(azks.get_root_hash::<_, Blake3>(&db).await?, &proof)?,
@@ -787,18 +842,18 @@ mod tests {
         let mut azks = Azks::new::<_, Blake3>(&db).await?;
 
         let mut insertion_set_1: Vec<Node<Blake3>> = vec![];
-        insertion_set_1.push(Node::<Blake3> {
-            label: NodeLabel::new(0b0, 64),
-            hash: Blake3::hash(&[]),
+        insertion_set_1.push(Node {
+            label: NodeLabel::new(byte_arr_from_u64(0b0), 64),
+            hash: Blake3::hash(&EMPTY_VALUE),
         });
         azks.batch_insert_leaves::<_, Blake3>(&db, insertion_set_1)
             .await?;
         let start_hash = azks.get_root_hash::<_, Blake3>(&db).await?;
 
         let mut insertion_set_2: Vec<Node<Blake3>> = vec![];
-        insertion_set_2.push(Node::<Blake3> {
-            label: NodeLabel::new(0b01 << 62, 64),
-            hash: Blake3::hash(&[]),
+        insertion_set_2.push(Node {
+            label: NodeLabel::new(byte_arr_from_u64(0b01 << 62), 64),
+            hash: Blake3::hash(&EMPTY_VALUE),
         });
 
         azks.batch_insert_leaves::<_, Blake3>(&db, insertion_set_2)
@@ -817,26 +872,27 @@ mod tests {
         let mut azks = Azks::new::<_, Blake3>(&db).await?;
 
         let mut insertion_set_1: Vec<Node<Blake3>> = vec![];
-        insertion_set_1.push(Node::<Blake3> {
-            label: NodeLabel::new(0b0, 64),
-            hash: Blake3::hash(&[]),
+        insertion_set_1.push(Node {
+            label: NodeLabel::new(byte_arr_from_u64(0b0), 64),
+            hash: Blake3::hash(&EMPTY_VALUE),
         });
-        insertion_set_1.push(Node::<Blake3> {
-            label: NodeLabel::new(0b1 << 63, 64),
-            hash: Blake3::hash(&[]),
+        insertion_set_1.push(Node {
+            label: NodeLabel::new(byte_arr_from_u64(0b1 << 63), 64),
+            hash: Blake3::hash(&EMPTY_VALUE),
         });
+
         azks.batch_insert_leaves::<_, Blake3>(&db, insertion_set_1)
             .await?;
         let start_hash = azks.get_root_hash::<_, Blake3>(&db).await?;
 
         let mut insertion_set_2: Vec<Node<Blake3>> = vec![];
-        insertion_set_2.push(Node::<Blake3> {
-            label: NodeLabel::new(0b01 << 62, 64),
-            hash: Blake3::hash(&[]),
+        insertion_set_2.push(Node {
+            label: NodeLabel::new(byte_arr_from_u64(0b1 << 62), 64),
+            hash: Blake3::hash(&EMPTY_VALUE),
         });
-        insertion_set_2.push(Node::<Blake3> {
-            label: NodeLabel::new(0b111 << 61, 64),
-            hash: Blake3::hash(&[]),
+        insertion_set_2.push(Node {
+            label: NodeLabel::new(byte_arr_from_u64(0b111 << 61), 64),
+            hash: Blake3::hash(&EMPTY_VALUE),
         });
 
         azks.batch_insert_leaves::<_, Blake3>(&db, insertion_set_2)
@@ -915,9 +971,8 @@ mod tests {
 
         let out = azks.get_root_hash_at_epoch::<_, Blake3>(&db, 123).await;
 
-        let expected = Err::<_, HistoryTreeNodeError>(HistoryTreeNodeError::NonexistentAtEpoch(
-            NodeLabel::root(),
-            0,
+        let expected = Err::<_, AkdError>(AkdError::HistoryTreeNode(
+            HistoryTreeNodeError::NonexistentAtEpoch(NodeLabel::root(), 123),
         ));
         assert_eq!(expected, out);
         Ok(())
