@@ -82,7 +82,7 @@ impl Transaction {
     }
 
     /// Start a transaction in the storage layer
-    pub async fn begin_transaction(&mut self) -> bool {
+    pub async fn begin_transaction(&self) -> bool {
         debug!("BEGIN begin transaction");
         let mut guard = self.state.write().await;
         let out = if (*guard).active {
@@ -96,18 +96,22 @@ impl Transaction {
     }
 
     /// Commit a transaction in the storage layer
-    pub async fn commit_transaction(&mut self) -> Result<Vec<DbRecord>, StorageError> {
+    pub async fn commit_transaction(&self) -> Result<Vec<DbRecord>, StorageError> {
         debug!("BEGIN commit transaction");
         let mut guard = self.state.write().await;
 
         if !(*guard).active {
-            return Err(StorageError::SetData(
+            return Err(StorageError::Transaction(
                 "Transaction not currently active".to_string(),
             ));
         }
 
         // copy all the updated values out
-        let records = guard.mods.values().cloned().collect();
+        let mut records = guard.mods.values().cloned().collect::<Vec<_>>();
+
+        // sort according to transaction priority
+        records.sort_by_key(|r| r.transaction_priority());
+
         // flush the trans log
         (*guard).mods.clear();
 
@@ -117,12 +121,12 @@ impl Transaction {
     }
 
     /// Rollback a transaction
-    pub async fn rollback_transaction(&mut self) -> Result<(), StorageError> {
+    pub async fn rollback_transaction(&self) -> Result<(), StorageError> {
         debug!("BEGIN rollback transaction");
         let mut guard = self.state.write().await;
 
         if !(*guard).active {
-            return Err(StorageError::SetData(
+            return Err(StorageError::Transaction(
                 "Transaction not currently active".to_string(),
             ));
         }
@@ -169,3 +173,87 @@ impl Transaction {
         debug!("END transaction set");
     }
 }
+
+#[cfg(test)]
+mod tests {
+    use super::*;
+    use crate::append_only_zks::*;
+    use crate::history_tree_node::*;
+    use crate::node_state::*;
+    use crate::storage::types::*;
+    use rand::{rngs::OsRng, seq::SliceRandom};
+
+    #[tokio::test]
+    async fn test_commit_order() -> Result<(), StorageError> {
+        let azks = DbRecord::Azks(Azks {
+            num_nodes: 0,
+            latest_epoch: 0,
+        });
+        let node1 = DbRecord::HistoryTreeNode(HistoryTreeNode {
+            label: NodeLabel::new(byte_arr_from_u64(0), 0),
+            birth_epoch: 1,
+            last_epoch: 1,
+            parent: NodeLabel::new(byte_arr_from_u64(0), 0),
+            node_type: NodeType::Root,
+        });
+        let node2 = DbRecord::HistoryTreeNode(HistoryTreeNode {
+            label: NodeLabel::new(byte_arr_from_u64(1), 1),
+            birth_epoch: 1,
+            last_epoch: 1,
+            parent: NodeLabel::new(byte_arr_from_u64(0), 0),
+            node_type: NodeType::Leaf,
+        });
+        let node_state1 = DbRecord::HistoryNodeState(HistoryNodeState {
+            value: [0u8; 32],
+            child_states: [None, None],
+            key: NodeStateKey(NodeLabel::new(byte_arr_from_u64(1), 1), 1),
+        });
+        let node_state2 = DbRecord::HistoryNodeState(HistoryNodeState {
+            value: [0u8; 32],
+            child_states: [None, None],
+            key: NodeStateKey(NodeLabel::new(byte_arr_from_u64(1), 1), 2),
+        });
+        let value1 = DbRecord::ValueState(ValueState {
+            username: AkdLabel::from_utf8_str("test"),
+            epoch: 1,
+            label: NodeLabel::new(byte_arr_from_u64(1), 1),
+            version: 1,
+            plaintext_val: AkdValue::from_utf8_str("abc123"),
+        });
+        let value2 = DbRecord::ValueState(ValueState {
+            username: AkdLabel::from_utf8_str("test"),
+            epoch: 2,
+            label: NodeLabel::new(byte_arr_from_u64(1), 1),
+            version: 2,
+            plaintext_val: AkdValue::from_utf8_str("abc1234"),
+        });
+
+        let records = vec![azks, node1, node2, node_state1, node_state2, value1, value2];
+        let mut rng = OsRng;
+
+        for _ in 1..10 {
+            let txn = Transaction::new();
+            txn.begin_transaction().await;
+
+            // set values in a random order
+            let mut shuffled = records.clone();
+            shuffled.shuffle(&mut rng);
+            for record in shuffled {
+                txn.set(&record).await;
+            }
+
+            // ensure that committed records are in ascending priority
+            let mut running_priority = 0;
+            for record in txn.commit_transaction().await? {
+                let priority = record.transaction_priority();
+                if priority > running_priority {
+                    running_priority = priority;
+                } else if priority < running_priority {
+                    panic!("Transaction did not obey record priority when committing");
+                }
+            }
+        }
+
+        Ok(())
+    }
+}
